Attribute VB_Name = "StrongHelp"
' Project: StrawHelp
' Module : StrongHelp
'
' Purpose: This module handles everything related to the access
'          and HTML conversion of a StrongHelp help file.
'
' Author : Rick Murray
'          heyrick1973 @ yahoo . co . uk
'          http://www.heyrick.co.uk/software/strawhelp/
'
' Version: 0.05
'
' Date   : 17th March 2015
' Started: 18th April 2008
'
' Licence:
'
'   StrawHelp is Copyright  2015 Rick Murray
'
'   Licensed under the EUPL, Version 1.1 ONLY (the "Licence");
'   You may not use this work except in compliance with the Licence.
'
'   You may obtain a copy of the Licence in all European languages at:
'     https://joinup.ec.europa.eu/software/page/eupl
'
'   A PDF version in English should have been supplied with this source.
'   If it was not, a copy may be found at the above URL.
'
'   Unless required by applicable law or agreed to in writing, software
'   distributed under the Licence is distributed on an "AS IS" basis,
'   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
'   implied. See the Licence for the specific language governing
'   permissions and limitations under the Licence.
'
'
'
' Points of contention:
'   1. #Include files are not supported. [##CHECK THIS##]
'   2. {$ <sysvar>) is ignored.
'   3. #Tab and alignment using tabs just doesn't work. There is
'      no clear way to make this work using HTML; which by
'      definition has omitted tabs from its eclectic repertoire.
'      Perhaps converting #Tab to a <table>? Hardcoded tabs will
'      be a little harder to support... if at all.
'   4. #Indent is 'faked' using blockquote and styles.
'   5. No sprites, no draw files.
'   6. Squashed content will result in a load of bollocks, as it
'      is only sane to squash sprites/drawfiles, and we don't do
'      either, there is thus no support for de-squashing.
'      [I'm hoping the Win32 decompress will so it, if SQSH is
'       basically a Unix LZW12 stream with a header attached...?]
'   7. Subpages are not supported (yet).
'   8. We don't support *bold* /italic/ _underline_ - you should
'      enclose them in {*}braces{*} for it to work.
'   9. This entry represents all the bugs and quirks that you will
'      encounter. So if you wish to complain about it showing the
'      document in "Hello Kitty" shades of pink (!), then I will
'      say "that was discussed under point #9". :-) :-)
'
' Note that navigation to a page will take you to that page. In
' StrongHelp it would open a NEW window, but I found that could
' get kinda messy to work with under the Windows metaphor.
'

Option Explicit


' External calls
' ==============

Private Declare Function IsAlphaNum Lib "User32" _
                  Alias "IsCharAlphaNumericA" _
                        (ByVal ch As Byte) As Boolean




' Array descriptions
' ==================

' StrongHelp file header block
Private Type StrHeaderDef
  ID As String * 4       ' will be "HELP"
  RootBlkSize As Long    ' will be 44 (or 40 with old software)
  CreatorVersion As Long ' is ignored
  OffsetToFree As Long   ' is ignored
End Type


' Directory definition header
Private Type DirBlockDef
  ID As String * 4       ' will be "DIR$"
  BlockSize As Long      ' is taken as size of directory data
  BlockUsed As Long      ' is (mostly) ignored
End Type
' Note that if (BlockSize == BlockUsed == 12)
' then the directory entry contains nothing.


' Directory entry definition (without object name), used in loader
Private Type DirEntryDataDef
  Offset As Long         ' offset from start of file, with start=0
  LoadAddr As Long       ' metadata: load address or file type
  ExecAddr As Long       ' metadata: exec address or datestamp
  Size As Long           ' file size in bytes
  Flags As Long          ' metadata: access/type flags (see consts)
  Reserved As Long       ' is ignored
End Type
' the string is read byte-by-byte in the loader, and
' the two parts are combined with other data to make:
Private Type DirEntryDef
  Offset As Long         ' offset from start of file (start = 0)
  LoadAddr As Long       ' load address / filetype
  ExecAddr As Long       ' exec address / datestamp
  Size As Long           ' size
  Flags As Long          ' access/type flags
  Reserved As Long       ' is ignored
  Name As String         ' object name
  
  ' these are my additions: easier to add into this array as it
  ' is being created than to have to work out each time.
  FQName As String       ' Fully Qualified (from '$') name
  BRName As String       ' Basic Reference: Full name with no '.'s
  Parent As Integer      ' Pointer to parent directory
  FileType As Integer    ' RISC OS file type (ZERO if untyped)
End Type


' File object header
Private Type FileDataDef
  ID As String * 4       ' will be "DATA"
  Size As Long           ' is ignored; recorded file size is used
End Type


' Included for completeness, we don't use this
Private Type FreeSpaceDef
  ID As String * 4       ' will be "FREE"
  Size As Long
  OffsetToNext As Long
End Type


' Font and style definition block ("one size fits all")
Private Type FontAttrDef
  ' These are used in font definitions
  Face As String           ' Font name ("Times New Roman" etc)
  Size As Integer          ' Font size, in points
  RefBold As Integer       ' Reference to 'bold' version
  RefItalic As Integer     ' Reference to 'italicised' version
  RefUnderline As Integer  ' Reference to 'underlined' version
  
  ' These are used in styles
  Name As String           ' The style name
  FontRef As Integer       ' The font to use; -1 for "no change"
  
  ' These are used in both
  Alignment As Integer     ' Alignment to use
  Bold As Boolean          ' TRUE if has BOLD attribute
  Italic As Boolean        ' TRUE if has ITALICISED attribute
  Underline As Boolean     ' TRUE if has UNDERLINED attribute
  Colour As Long           ' Colour (WIMP colour if < 16) else RGB
End Type

' Font alias table
Private Type FontAliasDef
  ROName As String
  HTMLName As String
End Type


' General document variables. Included in a Type so I can enter
' "DocVars." and as I press '.' VB will prompt me for which I
' would like to use. Better than trying to remember all of these!
Private Type DocVarsDef
  Link_Prefix As String      ' Prefix to add to links
  Link_Postfix As String     ' Postfix to suffix to links
  ParentDoc As String        ' Parent document reference
  Background As Long         ' Background colour in document
  DefaultBackground As Long  ' Default BG col from config
  Alignment As Integer       ' Current alignment
  AlignmentCache As Integer  ' Storage for previous alignment
  WrapMode As Integer        ' Current wrapping mode
  WrapModeCache As Integer   ' Storage for prev. wrap mode
  FontColourCache As Long    ' Storage for prev. font colour
End Type



' Now for the actual definitions
' ==============================

' This is a list of directory objects in the manual image file
Private Dirs() As DirEntryDef
Private DirCount As Integer


' This is a list of non-directory objects in the manual image file
Private Files() As DirEntryDef
Private FileCount As Integer


' This is the file name of the manual currently open
Private ManualName As String


' This is the current display style. '0' is the base style and
' apparently up to 32 levels of style can be used (has anybody?).
' The application of bold/italic/underlined are merged into the
' CURRENT style and upon reverting a level would be lost. Don't
' know if this is how StrongHelp does it, but it is how we are
' going to do it. One should close effects anyway, not dick around.
Private DispFont(0 To 32) As FontAttrDef
Private DispDepth As Integer


' The available font definitions (0-23 private; 24-31 user)
Private Font(0 To 31) As FontAttrDef


' Font aliases RISC OS -> HTML
Private FontAlias() As FontAliasDef
Private FontAliasCount As Integer


' The styles (we can hold up to 32 'styles'); StrongHelp says 16
' may be defined, and about 12 are set up automatically.
Private Style(1 To 32) As FontAttrDef
Private StyleCount As Integer


' Wimp colours
Private WimpColours(0 To 15) As Long


' For stashing output to be used in a tabled format.
Private TableData() As String
Private TableCount As Integer
Private InTable As Boolean
Private TableCols As Integer
Private TableRows As Integer


' History
Private PageStack() As Integer
Private StackCount As Integer


' Indentation
Private Indents() As Integer
Private IndentCount As Integer


' Internal page stuff
Private DocVars As DocVarsDef

Private DoingInclude As Boolean
Private IncludeFilePtr As Long
Private IncludePending As Boolean
Private IncludeReference As String

Private SubPageLevels As Integer


' Pointer to the document root.
Public ManualRootEntry As Integer



' Constants
' =========

' Directory object flags
' The first five equate to "RWL/rw" access flags. This is akin to
' "RW----rw-" access under Unix, only with a 'locked' addition.
' There is no useful equivalent under DOS/Windows, which is a bit
' of a joke, what with "system/hidden/archive/readonly"!
Private Const FLAG_OWNERREAD   As Long = 1
Private Const FLAG_OWNERWRITE  As Long = 2
' "Owner Execute Only" on BBC Master
' or "Private" on SJ file servers      = 4
Private Const FLAG_LOCKED      As Long = 8
Private Const FLAG_PUBLICREAD  As Long = 16
Private Const FLAG_PUBLICWRITE As Long = 32
' "Hidden" (Filer+)                    = 64
' "Application" (Filer+)
' ?"Hidden" on Level 4 FS?             = 128
' "Super Hidden" (Filer+) if 64+128...
Private Const FLAG_ISDIRECTORY As Long = 256

' Pointers to icons to use for files in the File System browser;
' these relate to the settings in the "imagelist"s.
Private Const ICON_DIRECTORY   As Integer = 1 ' directory (duh!)
' the following is only available as a "big" icon for the manuals
' listing; it is a copy of ICON_FILE_XXX in the small icons
Private Const ICON_FILE_3D6    As Integer = 2 ' StrongHelp file
Private Const ICON_FILE_AFF    As Integer = 3 ' DrawFile
Private Const ICON_FILE_FCA    As Integer = 4 ' Squashed file
Private Const ICON_FILE_FF9    As Integer = 5 ' SpriteFile
Private Const ICON_FILE_FFC    As Integer = 6 ' Code (for "!pre")
Private Const ICON_FILE_FFD    As Integer = 7 ' Data (non-search.)
Private Const ICON_FILE_FFF    As Integer = 8 ' Text (searchable)
Private Const ICON_FILE_XXX    As Integer = 9 ' Unknown (all else!)

' Alignments
Private Const ALIGN_LEFT       As Integer = 0 ' We actually JUSTIFY
Private Const ALIGN_CENTRE     As Integer = 1 ' Centred
Private Const ALIGN_RIGHT      As Integer = 2 ' Right-aligned

' Wrap options
' Wrap "Off" means when a newline is in the source file, it will
' be so in the displayed output.
Private Const WRAP_OFF         As Integer = 0
' Wrap "NoJoin" is much like wrap "Off", it expects each para to
' be one long line. Maybe "Wrap Off" goes off the right of the
' window? We cannot do that as it would look idiotic and not be
' very useful...
Private Const WRAP_NOJOIN      As Integer = 1
' Wrap on means a BLANK line is required for a paragraph, while
' newlines are suppressed and the text is flowed onto one para.
Private Const WRAP_ON          As Integer = 2



' Finally, some code!
' ===================

Public Sub Help_Initialise()
  ' The FIRST thing to be called, we should load the initial
  ' settings (like the global fonts, colours, etc)... We also
  ' look for manuals and update the Manuals form accordingly.
  ' No return code. A failure *here* is an abort.
  
  Dim MyFile As String, ManCount As Integer, InFile As String
  Dim IFP As Integer, OFP As Integer, OutFile As String
  Dim MyStr As String, Recurse As Integer
  
  ' Check we have manuals
  MyFile = Dir(App.Path & "\Manuals", vbDirectory)
  If (Len(MyFile) = 0) Then
    MsgBox "You must have a subdirectory within " & App.Title & _
           " called 'Manuals' which contains your StrongHelp " & _
           "manuals." & vbCrLf & "The default installation " & _
           "included a few manuals... did you delete these by " & _
           "mistake?", vbCritical + vbOKOnly
    End
  End If

  ' Enumerate manuals and update frmManuals
  Load frmManuals
  frmManuals.lvManuals.ListItems.Clear
  ' Look first for a literal filetype translation: blah.3d6
  MyFile = Dir(App.Path & "\Manuals\*.3d6")
  Do While (MyFile <> "")
    frmManuals.lvManuals.ListItems.Add , MyFile, _
                              Left$(MyFile, Len(MyFile) - 4), 1, 1
    ManCount = ManCount + 1
    MyFile = Dir
  Loop
  ' Look then for a nicer filetype translation: blah.shf
  ' shf = StrongHelp File
  MyFile = Dir(App.Path & "\Manuals\*.shf")
  Do While (MyFile <> "")
    frmManuals.lvManuals.ListItems.Add , MyFile, _
                              Left$(MyFile, Len(MyFile) - 4), 1, 1
    ManCount = ManCount + 1
    MyFile = Dir
  Loop

  If (ManCount = 0) Then
    MsgBox "You must have some manuals in the 'Manuals' " & _
           "subdirectory of " & App.Title & "..." & vbCrLf & _
           "The default installation included a few " & _
           "manuals... did you delete these by mistake?", _
           vbCritical + vbOKOnly
    End
  End If
  
  ' switch to small icons if there are more entries
  ' than will fit in 'normal' display style...
  If (ManCount > 28) Then
    frmManuals.lvManuals.View = lvwSmallIcon
  End If


  ' Load Wimp colours from "WimpColours.txt"
  ' This is specifically NOT embedded so the user can tweak stuff.
  IFP = FreeFile
  Open App.Path & "\WimpColours.txt" For Binary Access Read As IFP
  For Recurse = 0 To 15
    Line Input #IFP, MyStr
    WimpColours(Recurse) = Val("&H" & MyStr)
  Next
  Close #IFP


  ' Load font alias from "AliasFonts.txt"
  ' Format is:
  '   <RISC OS font>!<HTML font types>
  ' Like:
  '   Corpus!Courier New, mono
  ' (ditto re. tweaking stuff)
  IFP = FreeFile
  Open App.Path & "\FontAlias.txt" For Binary Access Read As IFP
  While (Loc(IFP) < LOF(IFP))
    Line Input #IFP, MyStr
    Recurse = InStr(1, MyStr, "!")
    FontAliasCount = FontAliasCount + 1
    ReDim Preserve FontAlias(FontAliasCount)
    FontAlias(FontAliasCount).ROName = Left$(MyStr, (Recurse - 1))
    FontAlias(FontAliasCount).HTMLName = Mid$(MyStr, (Recurse + 1))
  Wend
  Close #IFP

  
  ' Now fake a suitable enviroment to allow us to call the
  ' "Help_ScanForPreSettings()" function for the master
  ' "!Configure" file...
  InFile = App.Path & "\!Configure.txt"
  OutFile = App.Path & "\wdylast.txt"
  If (Len(Dir(InFile)) = 0) Then Exit Sub ' no !Config, drop out
  If (Len(Dir(OutFile)) <> 0) Then Kill OutFile

  IFP = FreeFile
  Open InFile For Binary Access Read As IFP
  OFP = FreeFile
  Open OutFile For Binary Access Write As OFP
  
  Help_ScanForPreSettings IFP, OFP

  Close #IFP
  Close #OFP
  
  Kill OutFile
  
  DocVars.DefaultBackground = DocVars.Background
End Sub


Public Function Help_Load(ByVal FileName As String) As Boolean
  ' Attempt to load the nominated manual. Returns TRUE if
  ' load was successful, else reports a reason and replies FALSE.
  
  Dim FP As Integer, Header As StrHeaderDef, MyObj As DirEntryDef
  Dim DirPtr As Long, DirID As DirBlockDef, DirSize As Long
  Dim SizeAdj As Long
  
  Help_Load = False
  DirCount = 0
  FileCount = 0
  ReDim Dirs(0)
  ReDim Files(0)
  
  ' Check file exists
  If (Len(Dir(FileName)) = 0) Then
    MsgBox "Sorry, the manual " & Chr$(34) & FileName & _
           Chr$(34) & " does not exist.", vbExclamation + vbOKOnly
    Exit Function
  End If
  
  ' Open file
  FP = FreeFile
  Open FileName For Binary Access Read As FP
  
  ' Load and validate header
  Get #FP, , Header
  
  If (Header.ID <> "HELP") Then
    Close #FP
    MsgBox "Sorry, this file does not appear to be a valid " & _
           "StrongHelp manual - the header ID is incorrect.", _
           vbExclamation + vbOKOnly
    Exit Function
  End If
  
  ' Correct size is 44, apparently versions before 2.60 said 40.
  If ((Header.RootBlkSize <> 40) And _
      (Header.RootBlkSize <> 44)) Then
    Close #FP
    MsgBox "Sorry, this file does not appear to be a valid " & _
           "StrongHelp manual - the header size is incorrect.", _
           vbExclamation + vbOKOnly
    Exit Function
  End If
  
  ' Get info on root directory
  MyObj = Help_LoadDirEntry(FP, 0)
  
  ' Validate the root directory locator data
  If (MyObj.Offset <> 44) Then
    ' ## Hack for AVR manual which is known to have root directory in an odd place
    If (MyObj.Offset <> 4688) Then
      MsgBox "Unexpected value for root directory pointer - " & _
             "was expecting 44, read " & CStr(MyObj.Offset) & _
             "; will continue but this may not work out..."
    End If
  End If
  
  If ((MyObj.Flags And FLAG_ISDIRECTORY) <> FLAG_ISDIRECTORY) Then
    Close #FP
    MsgBox "The root directory flags should say it is a " & _
           "directory (not a file), but they do not!", _
           vbExclamation + vbOKOnly
    Exit Function
  End If
  
  If (MyObj.Name <> "$") Then
    Close #FP
    MsgBox "The root directory should be " & Chr$(34) & "$" & _
           Chr$(34) & " but it is not, we read " & Chr$(34) & _
           MyObj.Name & Chr$(34) & ".", vbExclamation + vbOKOnly
    Exit Function
  End If
    
  ' Copy this into the directory structure
  DirCount = 1
  ReDim Dirs(DirCount)
  Dirs(DirCount) = MyObj
  
  ' Now we should build up a directory and file table.
  DirPtr = 1
  Do
    ' Go to the directory
    If ((Dirs(DirPtr).Offset + 1) > LOF(FP)) Then
      Close #FP
      MsgBox "Directory offset out of bounds, this manual " & _
             "is damaged and cannot be loaded...", _
             vbExclamation + vbOKOnly
      Exit Function
    End If
    Seek #FP, (Dirs(DirPtr).Offset + 1&)
    
    ' Load and check the directory header
    Get #FP, , DirID
    If (DirID.ID <> "DIR$") Then
      Close #FP
      MsgBox "Directory entry marker invalid at offset +" & _
             CStr(Dirs(DirPtr).Offset) & " for directory """ & _
             Dirs(DirPtr).FQName & """.", vbExclamation + vbOKOnly
      Exit Function
    End If
    
    ' If the block size is 12, it means the directory is EMPTY
    If (DirID.BlockSize > 12) Then
      ' size of directory
      DirSize = DirID.BlockUsed
      
      Do
        ' Load entry
        MyObj = Help_LoadDirEntry(FP, DirPtr)
        
        ' If it is a directory, add it...
        If ((MyObj.Flags And FLAG_ISDIRECTORY) = _
                             FLAG_ISDIRECTORY) Then
          ' add to the directory structure
          DirCount = DirCount + 1
          ReDim Preserve Dirs(DirCount)
          Dirs(DirCount) = MyObj
        Else
          ' add to the file structure
          FileCount = FileCount + 1
          ReDim Preserve Files(FileCount)
          Files(FileCount) = MyObj
        End If
      
        ' Reduce the directory size for object read
        SizeAdj = Len(MyObj.Name) + 1
        While ((SizeAdj Mod 4) <> 0)
          SizeAdj = SizeAdj + 1
        Wend
        DirSize = DirSize - (SizeAdj + 24)
        
      Loop Until (DirSize < 40) ' Smallest valid size: 12ID + 28Obj
    End If
    
    DirPtr = DirPtr + 1
  Loop Until (DirPtr > DirCount)
  
  ' Phew!
  Close #FP
  
  ManualRootEntry = Help_LocateByPath("$.!Root")
  
  ManualName = FileName
  
  Help_Load = True
End Function


Private Function Help_LoadDirEntry(ByVal FP As Integer, _
                         ByVal Parent As Integer, _
                Optional ByVal Offset As Long = 0) As DirEntryDef
  ' Code to load a directory entry. The complication here is that
  ' strings are NULL terminated, and aligned to the next WORD
  ' boundary (word = 32 bit); which requires special handling.
  '
  ' While we are here, we can also deal with our extended data.
  ' Setting it up the one time will save time later on.
  
  Dim OData As DirEntryDataDef, OBlock As DirEntryDef
  Dim MyByte As Byte, OName As String, SLen As Integer
  Dim Paths() As String, PathCount As Integer, PathPtr As Integer
  Dim FileType As Long
  
  ' Go to offset, if a specific offset was given
  If (Offset <> 0) Then
    Seek #FP, (Offset + 1&)
  End If
  
  ' Get the numerical fields
  Get #FP, , OData
  
  ' Read the (null terminated) string
  OName = ""
  Do
    Get #FP, , MyByte
    If (MyByte <> 0) Then OName = OName & Chr$(MyByte)
  Loop Until (MyByte = 0)
  
  ' How much has been read?
  SLen = Len(OName) + 1
  
  ' Read (and discard) to word align
  If ((SLen Mod 4) <> 0) Then
    Do
      Get #FP, , MyByte
      SLen = SLen + 1
    Loop Until ((SLen Mod 4) = 0)
  End If
  
  ' Now build the data structure
  OBlock.Offset = OData.Offset
  OBlock.LoadAddr = OData.LoadAddr
  OBlock.ExecAddr = OData.ExecAddr
  OBlock.Flags = OData.Flags
  OBlock.Size = OData.Size
  OBlock.Reserved = OData.Reserved
  OBlock.Name = OName
  
  ' Set up the parent pointer for the Fully Qualified and
  ' Basic Reference names
  OBlock.Parent = Parent
  If (Parent = 0) Then
    ' root
    OBlock.FQName = OName
    OBlock.BRName = ""
  Else
    ' we should unwind...
    PathPtr = Parent
    While (PathPtr <> 0)
      PathCount = PathCount + 1
      ReDim Preserve Paths(PathCount)
      Paths(PathCount) = Dirs(PathPtr).Name
      PathPtr = Dirs(PathPtr).Parent
    Wend
    ' we added them backwards, so work back UP the array...
    While (PathCount > 0)
      OBlock.FQName = OBlock.FQName & Paths(PathCount) & "."
      If (Left$(Paths(PathCount), 1) <> "[") Then
        ' omit [p].printf
        '      ^^^
        OBlock.BRName = OBlock.BRName & Paths(PathCount)
      End If
      PathCount = PathCount - 1
    Wend
    OBlock.FQName = OBlock.FQName & OName
    If (OBlock.BRName <> "") Then
      OBlock.BRName = Mid$(OBlock.BRName, 2) ' clip off '$'
    End If
    OBlock.BRName = OBlock.BRName & OName
  End If
  
  ' Set up the file types
  If ((OBlock.Flags And FLAG_ISDIRECTORY) = FLAG_ISDIRECTORY) Then
    OBlock.FileType = &H1000 ' always 'dir'
  Else
    If ((Not (OBlock.LoadAddr And &HFFF00000)) = 1048575) Then
      ' Typed file
      FileType = ((OBlock.LoadAddr And &HFFF00) \ 256)
    Else
      ' Untyped file
      FileType = 0
    End If
    OBlock.FileType = CInt(FileType)
  End If
  
  Help_LoadDirEntry = OBlock
End Function


Public Sub Help_PopulateDir()
  ' This will build the directory "tree" visible in the File
  ' System browser.
  ' We use "ICON_DIRECTORY" both times as that is all it can be.
  
  Dim ObjRef As Integer, PKey As String, FileType As Integer
  Dim MyNode As Node
  
  ' clear old
  frmFileSys.tvDirs.Nodes.Clear
  
  ' add root directory
  Set MyNode = frmFileSys.tvDirs.Nodes.Add(, , "$", "$", _
                                           ICON_DIRECTORY)
  MyNode.Tag = 1
  
  If (DirCount = 1) Then Exit Sub
  
  ' add all the others
  For ObjRef = 2 To DirCount
    ' get reference of parent
    PKey = Dirs(Dirs(ObjRef).Parent).FQName
    ' add it
    Set MyNode = frmFileSys.tvDirs.Nodes.Add(PKey, tvwChild, _
                      Dirs(ObjRef).FQName, Dirs(ObjRef).Name, _
                      ICON_DIRECTORY)
    MyNode.Tag = ObjRef
  Next
End Sub


Public Sub Help_PopulateFile(ByVal Parent As Integer)
  ' This will, given a pointer to the 'container directory', set
  ' up the list of files within that directory in the list view
  ' of the File System browser, plus set up the correct icon, etc.
  '
  ' NOTE: Unlike the Windows Filer, we do *NOT* show directories
  '       within the list of files.
  
  Dim ObjRef As Integer, ObjCount As Integer, FileIcon As Integer
  Dim MyListItem As ListItem
  
  ' clear old
  frmFileSys.lvFiles.ListItems.Clear
  
  ' scan for how many files are in this directory
  ObjCount = 0
  For ObjRef = 1 To FileCount
    If (Files(ObjRef).Parent = Parent) Then
      ObjCount = ObjCount + 1
    End If
  Next
    
  ' nothing?
  If (ObjCount = 0) Then Exit Sub
  
  ' add them all
  For ObjRef = 1 To FileCount
    If (Files(ObjRef).Parent = Parent) Then
      Select Case Files(ObjRef).FileType
        Case &HAFF: FileIcon = ICON_FILE_AFF
        Case &HFCA: FileIcon = ICON_FILE_FCA
        Case &HFF9: FileIcon = ICON_FILE_FF9
        Case &HFFC: FileIcon = ICON_FILE_FFC
        Case &HFFD: FileIcon = ICON_FILE_FFD
        Case &HFFF: FileIcon = ICON_FILE_FFF
        Case Else: FileIcon = ICON_FILE_XXX
      End Select
      Set MyListItem = frmFileSys.lvFiles.ListItems.Add(, _
        Files(ObjRef).FQName, Files(ObjRef).Name, _
        FileIcon, FileIcon)
      MyListItem.Tag = ObjRef
    End If
  Next
  
  ' force everything to be drawn side-by-side instead of with annoying gaps
  frmFileSys.lvFiles.Sorted = True
  frmFileSys.lvFiles.SortOrder = lvwDescending
  frmFileSys.lvFiles.Refresh
  frmFileSys.lvFiles.SortOrder = lvwAscending
  frmFileSys.lvFiles.Refresh
End Sub


Public Sub Help_OpenFile(ByVal Which As Integer, _
                         Optional ToNotepad As Boolean = False)
  ' Given a pointer to a file (in the files array), the file will
  ' be extracted to a separate file called "dataopen.txt" within
  ' StrawHelp. ## VISTA WON'T LIKE THIS :-) ##
  '
  ' This is currently used for loading help pages; they are
  ' extracted and them parsed individually.
  '
  ' The File System browser, however, would not display them but
  ' would instead open them in Notepad. So set <ToNotepad> to be
  ' TRUE to make this happen.
  
  Dim FileData() As Byte, FP As Integer, NewName As String
  Dim FileLen As Long, FilePos As Long
    
  FilePos = Files(Which).Offset + 9 ' skip over "DATA<size>" header
  FileLen = Files(Which).Size - 9   ' omit "DATA<size>" header + correction
  
  If (FileLen < 1) Then
    MsgBox "Unable to open file, file length is zero or invalid.", _
           vbExclamation + vbOKOnly
    Exit Sub
  End If
  
  FP = FreeFile
  Open ManualName For Binary Access Read As FP
  
  If (FilePos < 1) Then
    Close #FP
    MsgBox "Unable to open file, its pointer is before the " & _
           "beginning of the manual's image file...", _
           vbExclamation + vbOKOnly
    Exit Sub
  End If
  
  If (FilePos > LOF(FP)) Then
    Close #FP
    MsgBox "Unable to open file, its pointer is beyond the " & _
           "end of manual's image file...", _
           vbExclamation + vbOKOnly
    Exit Sub
  End If
  
  If ((FilePos + FileLen) > LOF(FP)) Then
    Close #FP
    MsgBox "Unable to open file, its extent reaches beyond " & _
           "the end of manual's image file...", _
           vbExclamation + vbOKOnly
    Exit Sub
  End If
  
  ' get the file
  ReDim FileData(FileLen)
  Seek #FP, FilePos
  Get #FP, , FileData
  
  Close #FP
  
  ' Now save it
  NewName = App.Path & "\dataopen.txt"
  If (Len(Dir(NewName)) <> 0) Then Kill NewName
  FP = FreeFile
  Open NewName For Binary Access Write As FP
  Put #FP, , FileData
  Close #FP
  
  If (ToNotepad <> False) Then
    Shell "Notepad " & Chr$(34) & NewName & Chr$(34), vbNormalFocus
  End If
End Sub


Public Function Help_LocateByAddress(ByVal Addr As String, _
                            Optional OnlyLoad As Boolean = False) _
                                     As Boolean
  ' Given a StrawHelp locator (with "strawhelp:" suffix clipped),
  ' this will try to load up the specified help page.
  ' Returns TRUE if failed (as True = Cancel load); else FALSE.
  
  ' We must be able to sort out the cocked-up mess that would
  ' result from stuff like:
  '   Wimp_PollIdle
  '      $.Wimp_PollIdle
  '      $.Wimp_.PollIdle
  '      $.Wimp_.Poll.Idle
  ' Or:
  '   printf
  '      $.p.rintf
  '      $.[p].printf
  '      $.[p].[p].[p].printf
  '
  ' **IMPORTANT** The OS320 manual root sets the POSTFIX to be "_"
  ' and then includes references to things like:
  '   <Image filesystems=>ImageEntry_>
  '   <Filesystems=>FSEntry_>
  ' which results in eventual links "ImageEntry__" and "FSEntry__"
  ' so we MUST be aware of a double underscore ending, and to
  ' reduce this to a single underscore.
  ' Is this a 'bug' in the OS320 manual, or is it somehow intended?
  '
  ' Another potential bug is a reference to an actual page which
  ' then has the ":!Root" suffix... Look in the "$.About" page.
  '
  ' What the F#@& (!) am I supposed to do with the FileCore_DiscOp
  ' page in the OS320 manual? Read it and weep:
  '   #Prefix FileCore_DiscOp_
  '   #Parent FileCore_
  '   # Gerph> I've reorganised this page, 'cos it looked 'awkward' before
  '   =>  R1 =  reasons and <flags>
  '             0 <Verify=>FileCore_DiscOp0>
  '             1 <Read sectors=>FileCore_DiscOp1>
  ' [etc]
  ' Yes, you guessed it, the eventual address turns out to be:
  '    FileCore_DiscOp_FileCore_DiscOp0 (etc)
  ' which is *complete* bollocks. It's a shame the StrongHelp
  ' software tries so hard to support this crap, else we'd have
  ' had a fairly simple filename parser...
  '
  
  Dim Object As Integer, MyAddr As String, ThisOne As Integer
  Dim MyAddrB As String
  
  Help_LocateByAddress = True
  MyAddr = Addr
  
  ' Remove "BlahBlah__" stuff that can be encountered with the
  ' OS320 manual that links to "BlahBlah_" with postfix "_".
  '                 ^^ double underscore up there,
  '                    'n case you missed it :-)
  If (Right$(MyAddr, 2) = "__") Then
    MyAddr = Left$(MyAddr, Len(MyAddr) - 1)
  End If
  
'  ' Remove the extraneous ":!Root" garbage that seems to be
'  ' rather prevalent in the OS320 manual; like in the "about"
'  ' page when it links to the various tables...
'  If (LCase$(Right$(MyAddr, 6)) = ":!root") Then
'    MyAddr = Left$(MyAddr, Len(MyAddr) - 6)
'  End If
  
  ' Fail subpage references
  If (Left$(MyAddr, 1) = ".") Then
    MsgBox "Sorry, sub-pages are not currently supported. " & _
           "scroll this document, it'll be down there " & _
           "somewhere...", vbExclamation + vbOKOnly
    Exit Function
  End If
  
  ' Fail offsite URLs
  If (LCase$(Left$(MyAddr, 4)) = "#url") Then
    MsgBox "Sorry, for reasons of security " & App.Title & _
           " will not attempt to open/access non-local " & _
           "addresses. Please load your preferred browser or " & _
           "email software (as appropriate)...", _
           vbExclamation + vbOKOnly
    Exit Function
  End If
  
  ' Fail links to other manuals (i.e. "Wimp:Wimp_PollIdle")
  Object = InStr(1, MyAddr, ":")
  If (Object <> 0) Then
    MsgBox "Sorry, cross-linking to other manuals is not " & _
           "currently supported, you will need to open the " & _
           Chr$(34) & Left$(MyAddr, (Object - 1)) & Chr$(34) & _
           " manual yourself..." & vbCrLf & Addr, _
           vbExclamation + vbOKOnly
    Exit Function
  End If
  
  ' See if we can find an "easy" match
  For ThisOne = 1 To FileCount
    If (StrComp(Files(ThisOne).BRName, MyAddr, _
                vbTextCompare) = 0) Then
      If (OnlyLoad = True) Then
        ' we only want to load it up [used for #Include]
        Help_OpenFile ThisOne, False
      Else
        ' Load it, convert it, show it
        Help_PushToStack ThisOne
        Help_ConvertToHTML ThisOne
      End If
      Help_LocateByAddress = False
      Exit Function
    End If
  Next
  
  ' Does it exist as a directory?
  For ThisOne = 1 To DirCount
    If (StrComp(Dirs(ThisOne).BRName, MyAddr, _
                vbTextCompare) = 0) Then
      ' Okay, work out an address for the !Root version
      ' and see if that exists...
      MyAddrB = Dirs(ThisOne).FQName & ".!root"
      Object = Help_LocateByPath(MyAddrB)
      If (Object <> 0) Then
        ' Found it!
        If (OnlyLoad = True) Then
          ' we only want to load it up [used for #Include]
          Help_OpenFile Object, False
        Else
          ' Load it, convert it, show it
          Help_PushToStack Object
          Help_ConvertToHTML Object
        End If
        Help_LocateByAddress = False
        Exit Function
      End If
    End If
  Next
  
  ' If we *still* cannot find it, if it ends with "_" then
  ' clip off the "_" and try *again*...
  ' OS320 manual -> main !Root to OS_Module_flags_ (from postfix).
  If (Right$(MyAddr, 1) = "_") Then
    MyAddr = Left$(MyAddr, (Len(MyAddr) - 1))
    If (Help_LocateByAddress(MyAddr) = False) Then
      Help_LocateByAddress = False
      Exit Function
    End If
  End If
  
  ' If we are still here (bored now!), then we will finally try
  ' InStr matching. We match in reverse, not just as the stepping
  ' backwards and also we look for the *real* name within *our*
  ' string, instead of the other way around.
  For ThisOne = FileCount To 1 Step -1
    If (InStr(1, MyAddr, Files(ThisOne).BRName, _
              vbTextCompare) <> 0) Then
      If (OnlyLoad = True) Then
        ' we only want to load it up [used for #Include]
        Help_OpenFile ThisOne, False
      Else
        ' Load it, convert it, show it
        Help_PushToStack ThisOne
        Help_ConvertToHTML ThisOne
      End If
      Help_LocateByAddress = False
      Exit Function
    End If
  Next
  ' Does it exist as a directory?
  For ThisOne = DirCount To 2 Step -1
    If (InStr(1, MyAddr, Dirs(ThisOne).BRName, _
              vbTextCompare) <> 0) Then
      ' Okay, work out an address for the !Root version
      ' and see if that exists...
      MyAddrB = Dirs(ThisOne).FQName & ".!root"
      Object = Help_LocateByPath(MyAddrB)
      If (Object <> 0) Then
        ' Found it!
        If (OnlyLoad = True) Then
          ' we only want to load it up [used for #Include]
          Help_OpenFile Object, False
        Else
          ' Load it, convert it, show it
          Help_PushToStack Object
          Help_ConvertToHTML Object
        End If
        Help_LocateByAddress = False
        Exit Function
      End If
    End If
  Next

  ' Sorry, can't figure this mess out but I tried really hard!...
  MsgBox "Sorry, can't resolve pointer to help page:" & vbCrLf & _
         vbTab & "Parsed" & vbTab & MyAddr & vbCrLf & vbTab & _
         "Original" & vbTab & Addr & vbCrLf & vbCrLf & _
         "Please contact <heyrick1973@yahoo.co.uk> describing " & _
         "which manual and file reference is causing this " & _
         "message to appear." & vbCrLf & "I may ask for a " & _
         "copy of the manual concerned, if I do not have a " & _
         "copy of it myself." & vbCrLf, vbExclamation + vbOKOnly
End Function


Public Function Help_LocateByPath(ByVal MyFile As String) _
                                  As Integer
  ' Given a fully qualified path (i.e. from '$'), case not of
  ' importance, this will return ZERO for a non-match, or else
  ' the file number of the one that matches.
  ' THIS DOES NOT MATCH DIRECTORIES.
  
  Dim Recurse As Integer
  
  Help_LocateByPath = 0
  
  For Recurse = 1 To FileCount
    If (StrComp(Files(Recurse).FQName, MyFile, _
                vbTextCompare) = 0) Then
      Help_LocateByPath = Recurse
      Exit Function
    End If
  Next
End Function


Public Sub Help_ConvertToHTML(ByVal Which As Integer)
  ' The grunt-work part of converting a StrongHelp document to
  ' as hopefully good an HTML version as we can manage.
  ' [though, tailored to look "correct" on the MSIE plugin
  '  rather than be actual decent HTML]
  '
  ' Calls helper functions, else this procedure would be mental.
  
  Dim InFile As String, OutFile As String, Title As String
  Dim IFP As Integer, OFP As Integer, MyLine As String
  Dim OutStr As String, MyByte As Byte, Recurse As Integer
  Dim MyPrev As Byte, TmpStr As String, ConfigFile As Integer
  
  ' set default state
  DocVars.Alignment = ALIGN_LEFT
  DocVars.AlignmentCache = ALIGN_LEFT
  DocVars.Background = DocVars.DefaultBackground
  DocVars.FontColourCache = 0          ' defaults to black
  DocVars.Link_Postfix = ""
  DocVars.Link_Prefix = ""
  DocVars.ParentDoc = ""
  DocVars.WrapMode = WRAP_OFF          ' defaults to NO wrapping
  DocVars.WrapModeCache = WRAP_OFF
  
  IndentCount = 0
  ReDim Indents(IndentCount)
  
  TableCount = 0
  ReDim TableData(TableCount)
  TableCols = -1
  TableRows = -1

  DoingInclude = False
  IncludeFilePtr = 0
  IncludePending = False
  IncludeReference = ""
  
  SubPageLevels = 0

  On Error GoTo ConvHTMLError

  ' set names - temp files ahoy!
  InFile = App.Path & "\dataopen.txt"
  OutFile = App.Path & "\strhlppage.html"
  
  ' make sure input and output files do NOT exist
  If (Len(Dir(InFile)) <> 0) Then Kill InFile
  If (Len(Dir(OutFile)) <> 0) Then Kill OutFile

  ' open output file
  OFP = FreeFile
  Open OutFile For Binary Access Write As OFP

  ' look FIRST to see if there is an !Configure file
  ConfigFile = Help_LocateByPath("$.!Configure")
  If (ConfigFile <> 0) Then
    ' Extract it and parse this FIRST
    Help_OpenFile ConfigFile
    If (Len(Dir(InFile)) <> 0) Then
'     TmpStr = "<!-- scanning local !Configure for " & _
               "definitions -->" & vbCrLf
'     Put #OFP, , TmpStr
      IFP = FreeFile
      Open InFile For Binary Access Read As IFP
      Help_ScanForPreSettings IFP, OFP
      Close #IFP
      Kill InFile
    End If
    ' if file couldn't be opened, just ignore it...
  End If

  ' extract the file we wanted
  Help_OpenFile Which
  
  ' does it exist?
  If (Len(Dir(InFile)) = 0) Then
    MsgBox "Conversion unsuccessful - data could not be " & _
           "extracted.", vbExclamation + vbOKOnly
    Exit Sub
  End If
  
  ' open input file
  IFP = FreeFile
  Open InFile For Binary Access Read As IFP
  
  ' Read through looking ONLY for "#Background" and "#f" with '='
' TmpStr = "<!-- scanning help file for definitions -->" & vbCrLf
' Put #OFP, , TmpStr
  Help_ScanForPreSettings IFP, OFP
  
  ' read the title
  Title = Help_ReadLine(IFP)
    
  ' write an HTML header, just claim we're HTML 3.2, IE won't care
  MyLine = "<!doctype html public " & Chr$(34) & "-//W3C//DTD " & _
           "HTML 3.2//EN" & Chr$(34) & ">" & vbCrLf & vbCrLf & _
           "<html lang=" & Chr$(34) & "en-gb" & Chr$(34) & ">" & _
           vbCrLf & "<head>" & vbCrLf & "<title>" & Title & _
           "</title>" & vbCrLf & "<base target=" & Chr$(34) & _
           "_top" & Chr$(34) & ">" & vbCrLf
  Put #OFP, , MyLine
  Help_WriteMeta OFP, "generator", App.Title & "/" & AppVers
  Help_WriteMeta OFP, "rating", "general"
  Help_WriteMeta OFP, "MSSmartTagsPreventParsing", "TRUE"
  Help_HTTPEquiv OFP, "content-type", "text/html; charset=iso-8859-1"
  Help_HTTPEquiv OFP, "content-language", "en"
  Help_HTTPEquiv OFP, "pragma", "no-cache"
  Help_HTTPEquiv OFP, "expires", "Thu, 01 Jan 1970 00:00:00 GMT"
  
  MyLine = "<style type=" & Chr$(34) & "text/css" & Chr$(34) & _
           ">" & vbCrLf & "<!--" & vbCrLf & "H1        { " & _
           "font-size: 220%; font-family: 'Courier New', " & _
           "mono; font-weight: bolder; border-bottom: " & _
           "thin dotted; line-height: 120%; }" & vbCrLf & _
           "ADDRESS   { color:#000033; font-size: 65%; " & _
           "font-family: 'Comic Sans MS', fantasy; border-" & _
           "top: thin dotted }" & vbCrLf & "A:link    { " & _
           "text-decoration: underline }" & vbCrLf & _
           "A:visited { text-decoration: underline }" & _
           vbCrLf & "A:hover   { text-decoration: none; " & _
           "font-weight: bolder }" & vbCrLf & "A:active  { " & _
           "text-decoration: none; font-weight: bolder }" & _
           vbCrLf & "-->" & vbCrLf & "</style>" & vbCrLf
  Put #OFP, , MyLine
  
  MyLine = "</head>" & vbCrLf & vbCrLf & _
           Help_ReturnBackground & "<a name=" & Chr$(34) & _
           "topofdoc" & Chr$(34) & ">" & vbCrLf & "<nobr><h1>" & _
           Help_HTMLSafe(Title) & "</h1></nobr>" & vbCrLf & _
           vbCrLf & "</a><div align=" & _
           Chr$(34) & "justify" & Chr$(34) & ">" & vbCrLf
  Put #OFP, , MyLine
  
  MyByte = 10
  While (Loc(IFP) < LOF(IFP))
    MyPrev = MyByte
    Get #IFP, , MyByte
    
    Select Case MyByte
      Case 10: ' newline
        OutStr = OutStr & vbCrLf
        Help_WriteLine OFP, OutStr
        
        If (InTable = False) Then
          ' only outside of tables; in table it is strictly WRAP OFF
          Select Case DocVars.WrapMode
            Case WRAP_OFF, WRAP_NOJOIN:
              If (MyPrev = 10) Then
                OutStr = "<p>" & vbCrLf
              Else
                OutStr = "<br>" & vbCrLf
              End If
              Help_WriteLine OFP, OutStr
          
            Case WRAP_ON:
              If (MyPrev = 10) Then
                OutStr = "<p>" & vbCrLf
                Help_WriteLine OFP, OutStr
              End If
          End Select
        End If
        OutStr = ""

      Case 13: ' omit
      
      Case 35: ' # - introduce a command like "#Line"
        If (MyPrev = 10) Then
          ' flush output
          Help_WriteLine OFP, OutStr
          TmpStr = Help_ReadLine(IFP)
          OutStr = Help_ParseCommand(TmpStr, OFP, True)
          MyByte = 10
          
          ' switching to another file?
          If (IncludePending = True) Then
            IncludePending = False
            
            Put #OFP, , OutStr
            OutStr = ""
            
            If (DoingInclude = True) Then
              ' Multiple inclusions are not possible
              MsgBox "Multiple levels of file inclusion are " & _
                     "not supported.", vbExclamation + vbOKOnly
              OutStr = OutStr & vbCrLf & "<p>" & vbCrLf & _
                       "[contents of " & Chr$(34) & _
                       IncludeReference & Chr$(34) & " here]" & _
                       vbCrLf & "<p>" & vbCrLf & vbCrLf
                       
            Else
              ' Include a new file
              ' Remember where we are and then rename
              ' the current file as "inc_<blah>".
              IncludeFilePtr = Seek(IFP)
              Close #IFP
              If (Len(Dir(InFile & ".inc")) <> 0) Then _
                Kill InFile & ".inc"
              Name InFile As InFile & ".inc"
              ' Extract the new file
              If (Help_LocateByAddress(IncludeReference, True) = _
                False) Then
                ' NOTE: FALSE = SUCCESS [flags for an icon...]
                Open InFile For Binary Access Read As IFP
                DoingInclude = True
              Else
                ' Can't open it; undo everything
                Name InFile & ".inc" As InFile
                Open InFile For Binary Access Read As IFP
                Seek #IFP, IncludeFilePtr
              End If ' did LocateByAddress work?
            End If ' multiple includes?
          End If ' include pending?
        End If ' myprev was ASCII 10?
        
      Case 38: ' & - convert for HTML
        OutStr = OutStr & "&amp;"
      
      Case 42: ' * - should we now toggle bold?
        OutStr = OutStr & Help_WonkAnEffect(MyByte, MyPrev, IFP)
      
      Case 47: ' / - should we now toggle italics?
        OutStr = OutStr & Help_WonkAnEffect(MyByte, MyPrev, IFP)

      Case 60: ' < - introduce link
        TmpStr = Help_ReadUntil(IFP, Asc(">"), True)
        If (Left$(TmpStr, 4) <> "&lt;") Then
          OutStr = OutStr & Help_ParseLink(TmpStr)
        Else
          OutStr = OutStr & TmpStr
        End If
        
      Case 62: ' > - convert for HTML
        OutStr = OutStr & "&gt;"
        
      Case 92: ' \ - escape the next character; convert if < or >
        Get #IFP, , MyByte
        Select Case MyByte
          Case 60: OutStr = OutStr & "&lt;"
          Case 62: OutStr = OutStr & "&gt;"
          Case Else: OutStr = OutStr & Chr$(MyByte)
        End Select
        
'      Case 95: ' _ - should we now toggle underlined?
'        OutStr = OutStr & Help_WonkAnEffect(MyByte, MyPrev, IFP)
'        ...uh... remind me again why this is disabled? ;-)
        
      Case 123: ' { - introduce in-line command
        TmpStr = Help_ReadUntil(IFP, Asc("}"), False)
        OutStr = OutStr & Help_ParseCommand(TmpStr, OFP, False)
        
      Case 143: ' RISC OS style bullet, convert to Symbol bullet
        OutStr = OutStr & "<font face=" & Chr$(34) & "Symbol" & _
                 Chr$(34) & ">" & Chr$(&HB7) & "</font>"

      Case Else: OutStr = OutStr & Chr$(MyByte)
    End Select
    
    ' catch EOFs within inclusions
    If (Loc(IFP) = LOF(IFP)) Then
      If (DoingInclude = True) Then
        ' restore original file
        Close #IFP
        Kill InFile
        Name InFile & ".inc" As InFile
        Open InFile For Binary Access Read As IFP
        Seek #IFP, IncludeFilePtr
        DoingInclude = False
      End If
    End If
  Wend

  ' catch EOF with no newline [the >8 is to trap any invalid read]
  If (Len(OutStr) > 8) Then
    Put #OFP, , OutStr
    OutStr = ""
  End If

  ' patch in a link to the top if 'subpages'
  If (SubPageLevels > 0) Then
    MyLine = "<a href=" & Chr$(34) & "#topofdoc" & Chr$(34) & _
             "><small>(return to the top)</small></a>" & _
             vbCrLf & "<p>" & vbCrLf & vbCrLf & vbCrLf
    Put #OFP, , MyLine
  End If

  ' footer
  MyLine = vbCrLf & "</div><p>" & vbCrLf & vbCrLf & vbCrLf & _
           "<address><span style=" & Chr$(34) & "font-style: " & _
           "normal" & Chr$(34) & ">Converted with " & App.Title & _
           " v" & AppVers & " (" & AppDate & ") by " & AppAuth & _
           " &nbsp;&nbsp;&nbsp;" & AppSupp & "<br>This software " & _
           "is licenced under the EUPL v1.1, refer to https://" & _
           "joinup.ec.europa.eu/software/page/eupl</span>" & _
           "</address>" & vbCrLf & "</body>" & vbCrLf & _
           "</html>" & vbCrLf
  Put #OFP, , MyLine
  
  ' done, tidy up
  Close #IFP
  Close #OFP
  
  ' kick it to the browser
  frmMain.Caption = "StrawHelp - " & Title
  frmMain.brwWebBrowser.Navigate OutFile
  frmMain.Show
  Exit Sub

ConvHTMLError:
  MsgBox "Parser error: " & Err.Description & vbCrLf & vbCrLf & _
         "Will try to continue going...", _
         vbExclamation + vbOKOnly
  Err.Clear
  Resume Next
End Sub


Private Sub Help_ScanForPreSettings(ByVal FP As Integer, _
                                    ByVal OFP As Integer)
  ' Given input/output pointers already defined, we look through
  ' the input (which may be a page OR an !Configure file) to see
  ' if there are background/font/style definitions.
  ' If there are, we act upon them.
  '
  Dim MyStr As String, PrefixHash As Boolean
  
  PrefixHash = False
  
  While (Loc(FP) < LOF(FP))
    MyStr = Help_ReadLine(FP)
    
    ' The "#Commands" command allows StrongHelp to omit all the
    ' extraneous '#' from lines, by specifying "these following
    ' lines are ALL commands".
    If (StrComp(MyStr, "#commands", vbTextCompare) = 0) Then
      PrefixHash = True
    End If
    
    ' If a hash is to be prefixed, do it now.
    If (PrefixHash = True) Then MyStr = "#" & MyStr
    
    ' Look for #Background ...
    If (LCase$(Left$(MyStr, 11)) = "#background") Then
      Help_ParseBackground Trim$(Mid$(MyStr, 12)), OFP
    
    ' Look for #F... = ...
    ElseIf (LCase$(Left$(MyStr, 2)) = "#f") Then
      If (InStr(1, MyStr, "=") <> 0) Then
        Help_ParseFontDef Mid$(MyStr, 3), OFP
      End If
    End If
  Wend
  
  ' Back to the beginning
  Seek #FP, 1
End Sub


Private Sub Help_WriteMeta(ByVal FP As Integer, _
                           ByVal MKey As String, _
                           ByVal MData As String)
  ' Write a meta tag. Saves cruddy code littering the place.
  ' Like: <meta name="rating" content="general">
  Dim MyLine As String
  
  MyLine = "<meta name=" & Chr$(34) & MKey & Chr$(34) & _
           " content=" & Chr$(34) & MData & Chr$(34) & ">" & _
           vbCrLf
  Put #FP, , MyLine
End Sub


Private Sub Help_HTTPEquiv(ByVal FP As Integer, _
                           ByVal MKey As String, _
                           ByVal MData As String)
  ' As above, but this time for the meta http-equiv stuff.
  ' Like: <meta http-equiv="content-language" content="en">
  Dim MyLine As String
  
  MyLine = "<meta http-equiv=" & Chr$(34) & MKey & Chr$(34) & _
           " content=" & Chr$(34) & MData & Chr$(34) & ">" & _
           vbCrLf
  Put #FP, , MyLine
End Sub


Private Sub Help_WriteLine(ByVal FP As Integer, _
                           ByVal MyLine As String)
  ' If NOT building a table, write directly to the file, else
  ' write into the table cache.
  
  If (InTable = False) Then
    Put #FP, , MyLine
  
  Else
    If (Right$(MyLine, 1) = Chr$(10)) Then
      MyLine = Left$(MyLine, Len(MyLine) - 2)
    End If
    TableCount = TableCount + 1
    ReDim Preserve TableData(TableCount)
    TableData(TableCount) = MyLine
  End If
End Sub


Private Function Help_HTMLSafe(ByVal MyStr As String) As String
  ' For inserting titles, we make sure it is HTML safe...
  
  Dim Posn As Integer, MyOut As String, MyByte As Byte
  
  If (Len(MyStr) = 0) Then Exit Function
  
  For Posn = 1 To Len(MyStr)
    MyByte = Asc(Mid$(MyStr, Posn, 1))
    
    Select Case MyByte
      Case 38: ' &
        MyOut = MyOut & "&amp;"
        
      Case 60: ' <
        MyOut = MyOut & "&lt;"
      
      Case 62: ' >
        MyOut = MyOut & "&gt;"
    
      Case Else:
        MyOut = MyOut & Chr$(MyByte)
    End Select
  Next
  
  Help_HTMLSafe = MyOut
End Function


Private Function Help_ReadLine(ByVal FP As Integer) As String
  ' Read a line of input from the file.
  ' We cannot use:
  '   Line Input #FP, MyLine
  ' because these files will be in RISC OS format, which uses
  ' a single 0x0A byte for line termination; unlike certain
  ' operating systems that need a wasteful CRLF pair.
  
  Dim MyInput As String, Recurse As String, FiddlePos As Long
  Dim MyByte As Byte, Offset As Long
  
  MyInput = ""
  MyByte = 0
  Do While (MyByte <> 10)
    Get #FP, , MyByte
      
    ' Clip out DOS style CRs (shouldn't happen, but...)
    If (MyByte = 13) Then MyByte = 0
      
    ' Append it, byte by byte, yawn
    If (MyByte <> 0) And (MyByte <> 10) Then
      MyInput = MyInput & Chr$(MyByte) ' does this screw with VB's stack?
    End If                             ' does anybody care?
      
    ' Trap EOF state
    If (Loc(FP) >= LOF(FP)) Then
      ' don't bother to throw an error, it'll get seen eventually
      Help_ReadLine = MyInput
      Exit Function
    End If
  Loop
    
  Help_ReadLine = MyInput
End Function


Private Function Help_ReadUntil(ByVal FP As Integer, _
                                ByVal MyChar As Byte, _
                                ByVal IsALink As Boolean) As String
  Dim TmpStr As String, MyByte As Byte, MyPrev As Byte
  
  MyByte = 0
  TmpStr = ""
  Do
    MyPrev = MyByte
    Get #FP, , MyByte
    
    If ((TmpStr = "") And (IsALink = True)) Then
      ' If if is "<=" or "<-" or "<<" then these are allowed to
      ' be included in the output file without being escaped.
      ' So many frickin' exceptions and weirdness that this
      ' parser's gonna be a bit of a mess.
      If ((MyByte = 61) Or _
          (MyByte = 45) Or _
          (MyByte = 60)) Then
        TmpStr = "&lt;" & Chr$(MyByte)
        Help_ReadUntil = TmpStr
        Exit Function
      End If
    End If

    If (IsALink = True) Then
      ' fudge around Blah=>Blah style links
      If (MyPrev = 61) And (MyByte = 62) Then MyByte = 127
    End If
    
    ' escaped stuff?
    If (MyByte = 92) Then
      Get #FP, , MyByte
      If (MyByte = 60) Then
        TmpStr = TmpStr & "&lt;"
      ElseIf (MyByte = 62) Then
        TmpStr = TmpStr & "&gt;"
      Else
        TmpStr = TmpStr & Chr$(MyByte)
      End If
      MyByte = 0
    End If
    
    ' if not the byte we're looking for, add it to the output
    If (MyByte <> MyChar) Then TmpStr = TmpStr & Chr$(MyByte)
  Loop Until (MyByte = MyChar)
  
  Help_ReadUntil = TmpStr
End Function


Private Function Help_ParseCommand(ByVal Cmnd As String, _
                                   ByVal FP As Integer, _
                                   ByVal FromHash As Boolean) _
                                   As String
  ' Parse a command. Note that we will receive commands from
  ' "#<command>" as well as "{<command>}". The wrapper (the '#'
  ' or the '{' and '}') are stripped so they will appear equal;
  ' use the FromHash bool for cases where it matters which...
  ' Note also that semi-colons separate multiple commands, so we
  ' might encounter "#Align Centre; Line 50; Align". And note
  ' finally that the font selection can be followed by a colon and
  ' THEN something to display in that font (and only in that font,
  ' it will auto-revert afterwards). StrongHelp is awesome powerful
  ' but it sucks parsing all these permutations! ;-)
  '
  '
  ' Those commands with '~' to the left are NOT supported.
  ' Those commands with '#' to the left are handled elsewhere.
  ' Those commands with '!' to the left are incomplete.
  ' Those commands with 'X' to the left are "to be done" (some day).
  '
  '   Comment        #<space>blahblah
  '   Bold on/off    *
  '   Ital. on/off   /
  '   Und. on/off    __  (actually only '_' but VB messes that up)
  '   Alignment      Align [Left|Centre|Right]
  ' # BG colour      Background <WimpCol>|RGB <r>,<g>,<b>|Tile <sp>
  ' ~ Force below    Below
  ' ~ Mark bottom    Bottom
  ' ~ Add DrawFile   Draw <x>|centre, <y>|float <filename>
  '   Table list     Table Columns|Lines <number>
  '   End of table   EndTable
  '!  Select font    "f" or "f<num>" or "f<name>"
  '!  Select font    Also "f*<num>" or "f/<num>" or "f_<num>"!
  '                  We must also support "#f<spec>:some text..."
  '!# Define font    f<num> = <font> <xsize> [<ysize>] <b> <i> <bt>
  '!# Define style   f<name> = [f[*][/][_]<num>] [rgb] [align]
  ' ~ Insert file    Include <filename>
  '   Indent         [+][<number>]
  '   Draw sep. line Line [<width %ge>]
  ' ~ Manuals        Insert list of manuals, is IGNORED
  'X  Parent         Specify parent
  ' ~ INTERNAL USE   Pointer [0|1]  - for internal use, is ignored
  '   Link prefix    Prefix <string>
  '   Link suffix    Postfix <string>
  'X  Set text col.  RGB [<red value>,<green value>,<blue value>]
  ' ~ Add a sprite   Sprite <x>|centre, <y>|float [,<width>,
  '                    <height>] <spritename> [<label>] [=>link]
  ' ~ Def spritefile SpriteFile <filename>
  '   Def subpage    Subpage <name>
  'X  Tabbed area    Tab [<format, like for font def.>]
  ' ~ Bookmark       Tag <label>
  '   Wrap options   Wrap [On|Off|NoJoin]
  '
  '
  ' NOTE: The file pointer passed is *ONLY* for use by the call
  '       to Help_CreateTable and Help_CreateTabbed as it would
  '       be way too unweildly to do otherwise...
  '       In ALL other cases, we should return our data in a
  '       formatted string ready to be written to the file...
  
  Dim MyLine As String, MyCmnd As String, Posn As Integer
  
  ' Trap comments
  If (Left$(Cmnd, 1) = " ") Then
    Help_ParseCommand = ""
    Exit Function
  End If
  
  ' copy the command string so we can mess with it
  MyCmnd = Trim$(Cmnd)
  If (FromHash = True) Then
    MyLine = vbCrLf & "<!-- hash command: " & Chr$(34) & MyCmnd & _
             Chr$(34) & "-->" & vbCrLf
  Else
    MyLine = "<!-- inline command: " & Chr$(34) & MyCmnd & _
             Chr$(34) & "-->"
  End If
  
  Do
    ' TEXTUAL EFFECTS
    ' ---------------
    If (Left$(MyCmnd, 1) = "*") Then
      ' Bold
      If (DispFont(DispDepth).Bold = False) Then
        MyLine = MyLine & "<b>"
        DispFont(DispDepth).Bold = True
      Else
        MyLine = MyLine & "</b>"
        DispFont(DispDepth).Bold = False
      End If
    
    ElseIf (Left$(MyCmnd, 1) = "/") Then
      ' Italicised
      If (DispFont(DispDepth).Italic = False) Then
        MyLine = MyLine & "<i>"
        DispFont(DispDepth).Italic = True
      Else
        MyLine = MyLine & "</i>"
        DispFont(DispDepth).Italic = False
      End If
    
    ElseIf (Left$(MyCmnd, 1) = "_") Then
      ' Underlined
      If (DispFont(DispDepth).Underline = False) Then
        MyLine = MyLine & "<u>"
        DispFont(DispDepth).Underline = True
      Else
        MyLine = MyLine & "</u>"
        DispFont(DispDepth).Underline = False
      End If
      
      
    ' ALIGNMENT
    ' ---------
    ElseIf (LCase$(Left$(MyCmnd, 5)) = "align") Then
      ' Align [Left|Centre|Right]
      MyCmnd = Trim$(Mid$(MyCmnd, 6))
      
      If (LCase$(Left$(MyCmnd, 4)) = "left") Then
        ' Left alignment
        MyLine = MyLine & "</div>" & vbCrLf & "<div align=" & _
                 Chr$(34) & "justify" & Chr$(34) & ">"
        DocVars.AlignmentCache = DocVars.Alignment
        DocVars.Alignment = ALIGN_LEFT
        ' ??                                                     ??
        ' ?? Should this be LEFT or JUSTIFY? Which looks better? ??
        ' ??                                                     ??
        
      ElseIf (LCase$(Left$(MyCmnd, 6)) = "centre") Then
        ' Centred
        MyLine = MyLine & "</div>" & vbCrLf & "<div align=" & _
                 Chr$(34) & "center" & Chr$(34) & ">"
        DocVars.AlignmentCache = DocVars.Alignment
        DocVars.Alignment = ALIGN_CENTRE
  
      ElseIf (LCase$(Left$(MyCmnd, 6)) = "center") Then
        ' Centred (with horrid American spelling)
        MyLine = MyLine & "</div>" & vbCrLf & "<div align=" & _
                 Chr$(34) & "center" & Chr$(34) & ">"
        DocVars.AlignmentCache = DocVars.Alignment
        DocVars.Alignment = ALIGN_CENTRE
  
      ElseIf (LCase$(Left$(MyCmnd, 5)) = "right") Then
        ' Right aligned
        MyLine = MyLine & "</div>" & vbCrLf & "<div align=" & _
                 Chr$(34) & "right" & Chr$(34) & ">"
        DocVars.AlignmentCache = DocVars.Alignment
        DocVars.Alignment = ALIGN_RIGHT
        
      Else
        ' Restore previous alignment
        Select Case DocVars.AlignmentCache
          Case ALIGN_LEFT:
            MyLine = MyLine & "</div>" & vbCrLf & "<div align=" & _
                     Chr$(34) & "justify" & Chr$(34) & ">"
          
          Case ALIGN_CENTRE:
            MyLine = MyLine & "</div>" & vbCrLf & "<div align=" & _
                     Chr$(34) & "center" & Chr$(34) & ">"
          
          Case ALIGN_RIGHT:
            MyLine = MyLine & "</div>" & vbCrLf & "<div align=" & _
                     Chr$(34) & "right" & Chr$(34) & ">"
        End Select
        DocVars.Alignment = DocVars.AlignmentCache
        DocVars.AlignmentCache = ALIGN_LEFT ' in case called twice
      End If
      
      
    ' FONT
    ' ----
    ElseIf (LCase$(Left$(MyCmnd, 1)) = "f") Then
      ' f<lots> - the font/style definition syntax is disgusting
      '           it is trying to do too much with one command...
      MyLine = Help_SortOutFontMess(Mid$(MyCmnd, 2), FromHash)
      If (InStr(1, MyCmnd, ";") = 0) Then MyCmnd = ""
      
      
    ' DRAW
    ' ----
    ElseIf (LCase$(Left$(MyCmnd, 4)) = "draw") Then
      ' Draw <x>, <y>, <file>
      '   X can be absolute position, or "centre" ("c")
      '   Y can be absolute position, or "float" ("f")
      If (LCase$(Mid$(MyCmnd, 6, 1)) = "c") Then
        MyLine = MyLine & vbCrLf & "<div align=""center""><font face" & _
                 "=""Verdana"">[ DrawFile illustration ]</font></div>" & vbCrLf
      Else
        MyLine = MyLine & vbCrLf & "<div align=""left""><font face" & _
                 "=""Verdana"">[ DrawFile illustration ]</font></div>" & vbCrLf
      End If
      MyCmnd = ""
      
      
    ' ENDTABLE
    ' --------
    ElseIf (LCase$(Left$(MyCmnd, 8)) = "endtable") Then
      ' EndTable
      InTable = False
      Help_CreateTable FP
      MyCmnd = ""
      
      
    ' INCLUDE
    ' -------
    ElseIf (LCase$(Left$(MyCmnd, 7)) = "include") Then
      ' Include <file>
      IncludeReference = Trim$(Mid$(MyCmnd, 8))
      IncludePending = True
      
      
    ' INDENT
    ' ------
    ElseIf (LCase$(Left$(MyCmnd, 6)) = "indent") Then
      ' Indent [<amount>]
      MyCmnd = Trim$(Mid$(MyCmnd, 7))
      If ((MyCmnd <> "") And (Left$(MyCmnd, 1) <> ";")) Then
        ' Indent with a value
        Posn = Val(MyCmnd)
        If (Posn = 0) Then
          ' support sanely specifying "Indent 0"
          If (IndentCount > 0) Then
            ' if there is an indent stack, destroy it
            IndentCount = 0
            ReDim Indents(IndentCount)
            MyLine = MyLine & "</blockquote>" & vbCrLf
          End If
        Else
          ' remove previous indent before setting this one?
          If (IndentCount > 0) Then
            MyLine = MyLine & "</blockquote>" & vbCrLf
          End If
          
          ' set up an indent stack
          Posn = Posn * 8 ' OS chars -> pixels
          IndentCount = IndentCount + 1
          ReDim Preserve Indents(IndentCount)
          Indents(IndentCount) = Posn
          MyLine = MyLine & "<blockquote style=" & Chr$(34) & _
                   "margin-top: 0px; margin-bottom: 0px; " & _
                   "margin-right: 0px; margin-left: " & _
                   CStr(Posn) & "px;" & Chr$(34) & ">"
        End If
      Else
        ' Indent alone
        If (IndentCount > 0) Then
          ' existing indent, remove it
          MyLine = MyLine & "</blockquote>" & vbCrLf
          IndentCount = IndentCount - 1
          ReDim Preserve Indents(IndentCount)
          If (IndentCount > 0) Then
            ' still an indent level, so set it...
            Posn = Indents(IndentCount)
            MyLine = MyLine & "<blockquote style=" & Chr$(34) & _
                     "margin-top: 0px; margin-bottom: 0px; " & _
                     "margin-right: 0px; margin-left: " & _
                     CStr(Posn) & "px;" & Chr$(34) & ">"
          End If
        End If
      End If
 
 
    ' LINE
    ' ----
    ElseIf (LCase$(Left$(MyCmnd, 4)) = "line") Then
      ' Line [<width>]    ### WIDTH NOT CURRENTLY SUPPORTED ###
      MyLine = MyLine & vbCrLf & "<hr size=2>" & vbCrLf & vbCrLf
      
      
    ' LINK PREFIX
    ' -----------
    ElseIf (LCase$(Left$(MyCmnd, 6)) = "prefix") Then
      ' Prefix <link prefix>
      DocVars.Link_Prefix = Trim$(Mid$(MyCmnd, 7))
      MyCmnd = "" ' assume everything is the prefix (valid?)

    ' LINK SUFFIX
    ' -----------
    ElseIf (LCase$(Left$(MyCmnd, 7)) = "postfix") Then
      ' Postfix <link suffix>
      DocVars.Link_Postfix = Trim$(Mid$(MyCmnd, 8))
      MyCmnd = "" ' assume everything is the suffix (valid?)


    ' MANUALS
    ' -------
    ElseIf (LCase$(Left$(MyCmnd, 7)) = "manuals") Then
      MyLine = MyLine & "<p>" & vbCrLf & "<h3 align=center>" & _
               "Your manuals cannot be listed in this way.<br>" & _
               "Please click on '<font face=Arial>Manuals " & _
               "<u>i</u>ndex</font>' to select a manual.</h3>" & _
               vbCrLf & "<p>" & vbCrLf & vbCrLf


    ' Sprite
    ' ------
    ElseIf (LCase$(Left$(MyCmnd, 6)) = "sprite") Then
      ' Sprite <x>, <y>, [<w>, <h>,] <file> <link> <some-other-stuff...>
      '   X can be absolute position, or "centre" ("c")
      '   Y can be absolute position, or "float" ("f")
      If (LCase$(Mid$(MyCmnd, 8, 1)) = "c") Then
        MyLine = MyLine & vbCrLf & "<div align=""center""><font face" & _
                 "=""Verdana"">[ Bitmap illustration ]</font></div>" & vbCrLf
      Else
        MyLine = MyLine & vbCrLf & "<div align=""left""><font face" & _
                 "=""Verdana"">[ Bitmap illustration ]</font></div>" & vbCrLf
      End If
      MyCmnd = ""
    
    
    ' SUBPAGE - just chuck in a separator and <a name> link...
    ' -------
    ElseIf (LCase$(Left$(MyCmnd, 7)) = "subpage") Then
      ' SubPage <name>
      MyCmnd = Trim$(Mid$(MyCmnd, 8))
      MyLine = MyLine & "<p>&nbsp;<p>" & vbCrLf & vbCrLf
      If (SubPageLevels > 0) Then
        MyLine = MyLine & "<a href=" & Chr$(34) & "#topofdoc" & _
                 Chr$(34) & "><small>(return to the top)" & _
                 "</small></a>" & vbCrLf & "<p>" & vbCrLf & _
                 vbCrLf & vbCrLf
      End If
      SubPageLevels = SubPageLevels + 1
      MyLine = MyLine & "<a name=" & Chr$(34) & MyCmnd & _
               Chr$(34) & "></a>" & vbCrLf & "<span style=" & _
               Chr$(34) & "background-color: red" & Chr$(34) & _
               "><font color=white face=" & Chr$(34) & _
               "Arial, sans-serif" & Chr$(34) & "><b>&nbsp;Sub" & _
               "page:&nbsp;</b></font></span>" & vbCrLf & "<p>" & _
               vbCrLf & vbCrLf
      MyCmnd = "" ' assume everything is the suffix (valid?)


    ' TABLE (start)
    ' -----
    ElseIf (LCase$(Left$(MyCmnd, 5)) = "table") Then
      ' Table [Columns | Lines] <number>
      MyCmnd = Trim$(Mid$(MyCmnd, 6))
 
      If (LCase$(Left$(MyCmnd, 7)) = "columns") Then
        ' By columns
        MyCmnd = Trim$(Mid$(MyCmnd, 8))
        TableCols = Val(MyCmnd)
        TableRows = -1
        If (TableCols > 0) Then
          InTable = True
        Else
          TableCols = -1
        End If
      ElseIf (LCase$(Left$(MyCmnd, 5)) = "lines") Then
        ' By rows
        MyCmnd = Trim$(Mid$(MyCmnd, 6))
        TableRows = Val(MyCmnd)
        TableCols = -1
        If (TableRows > 0) Then
          InTable = True
        Else
          TableRows = -1
        End If
      Else
        ' We are supposed to 'assume' columns...
        TableCols = Val(MyCmnd)
        If (TableCols > 16) Then
          ' Probably the odd #Table <num> <num> format?
          ' Seems this defines ROWS, not columns.
          Posn = InStr(1, MyCmnd, " ")
          TableRows = Val(Left$(MyCmnd, (Posn - 1)))
          TableCols = -1
        Else
          TableRows = -1
        End If
        If ((TableCols > 0) Or (TableRows > 0)) Then
          InTable = True
        Else
          TableCols = -1
        End If
      End If
      MyCmnd = ""


    ' WRAP
    ' ----
    ElseIf (LCase$(Left$(MyCmnd, 4)) = "wrap") Then
      ' Wrap [On|NoJoin|Off]
      MyCmnd = Trim$(Mid$(MyCmnd, 5))
      
      If (LCase$(Left$(MyCmnd, 2)) = "on") Then
        ' Wrap on
        DocVars.WrapModeCache = DocVars.WrapMode
        DocVars.WrapMode = WRAP_ON
        
      ElseIf (LCase$(Left$(MyCmnd, 6)) = "nojoin") Then
        ' Wrap NoJoin
        DocVars.WrapModeCache = DocVars.WrapMode
        DocVars.WrapMode = WRAP_NOJOIN
  
      ElseIf (LCase$(Left$(MyCmnd, 3)) = "off") Then
        ' Wrap Off
        DocVars.WrapModeCache = DocVars.WrapMode
        DocVars.WrapMode = WRAP_OFF
          
      Else
        ' Restore previous wrap
        DocVars.WrapMode = DocVars.WrapModeCache
        DocVars.WrapModeCache = WRAP_OFF
      End If


    ' Everything else...
    ' ------------------
    Else
      ' MyLine = "[COMMAND: " & Chr$(34) & MyCmnd & Chr$(34) & _
               "]" & vbCrLf
    End If
  
    Posn = InStr(1, MyCmnd, ";")
    If (Posn <> 0) Then
      MyCmnd = Trim$(Mid$(MyCmnd, (Posn + 1)))
    Else
      ' a bug in Gareth Boden's C manual, or an
      ' undocumented alternative method?
      Posn = InStr(1, MyCmnd, ",")
      If (Posn <> 0) Then
        MyCmnd = Trim$(Mid$(MyCmnd, (Posn + 1)))
      End If
    End If
  Loop Until (Posn = 0)
  
  Help_ParseCommand = MyLine
End Function


Private Function Help_ParseLink(ByVal Link As String) As String
  ' When we click on a link in the browser, if the link begins
  ' with "strawhelp:" then it will be passed to us. We should
  ' work out which page this link refers to, and then we should
  ' display it...
  Dim MyLine As String, Posn As Integer, LP As String, LS As String
  Dim LinkText As String, LinkAddr As String
  
  ' The browser control picks up on "strawhelp:" as a method and
  ' passes that back to us to lookup/load a new help page...
  
  ' Links are in the form:
  '   "blah"
  '     where 'blah' is both the link pointer and the text that
  '     is to be displayed for this link.
  '   "blah=[127]yack" ([127] = ASCII 127)
  '     where 'blah' is the text to be displayed, but 'yack' is
  '     where we should actually link to. this allows us to be
  '     tidy, like "<DragAnObject=>Wimp:DragAnObject_>" or to
  '     implement spaces in the display text, which would not be
  '     legal in filenames, like "<OS_Byte values=>OS_ByteIndex>".
  
  ' Copy the Prefix and Suffix
  LP = DocVars.Link_Prefix
  LS = DocVars.Link_Postfix
  
  ' Is link simple or complex?
  Posn = InStr(1, Link, Chr$(127))
  If (Posn = 0) Then
    ' Simple link, like "<ADFS>" where link and text are the same
    LinkText = Link
    LinkAddr = Link
  
  Else
    ' Complex link, like "<Wimp=>Wimp:Wimp_>" where the text is
    ' given in simple form (i.e. "Wimp") and the link is more
    ' complicated (i.e. "Wimp:Wimp_").
    LinkText = Left$(Link, (Posn - 2))
    LinkAddr = Mid$(Link, (Posn + 1))
  End If
  
  ' If the link address begins '.' or has a ':', destroy pref/suf.
  If ((InStr(1, LinkAddr, ":") <> 0) Or _
      (Left$(LinkAddr, 1) = ".")) Then
    ' get rid of the prefix and suffix for this link as it is
    ' either to a different help document or it is to a link...
    LP = ""
    LS = ""
  End If
  
  ' make the complete link address
  If (Left$(LinkAddr, 1) = ".") Then
    ' subpage link
    LinkAddr = "#" & Mid$(LinkAddr, 2)
  Else
    ' full link
    LinkAddr = "strawhelp:" & LP & LinkAddr & LS
  End If
  
  ' now make the correct HTML link
  MyLine = "<a href=" & Chr$(34) & LinkAddr & Chr$(34) & ">" & _
           LinkText & "</a>"
           
  ' return this, finally...
  Help_ParseLink = MyLine
End Function


Private Sub Help_ParseBackground(ByVal Cmnd As String, _
                                 ByVal FP As Integer)
  ' Called before HTML output, for "#Background" stuff. NOTE THAT
  ' THE "#BACKGROUND" IS CLIPPED, AS WE ALREADY *KNOW* THIS...
  '
  ' Our possibilities are:
  '
  '   #Background RBG <red>, <green>, <blue>
  '   #Background Wimp <num>
  '   #Background Tile <sprite>        [NOT SUPPORTED]
  '
  ' Prior to our call, the HTML convertor will have assigned the
  ' 'default' background to be our document background. This means
  ' the initial state is that as defined in the !Configure file.
  '
  ' We set the DocVars background as appropriate. Refer to the
  ' "Help_returnBackground()" function for specifics.
  '

  Dim MyStr As String, Posn As Integer, CmndCpy As String
  Dim Red As Integer, Green As Integer, Blue As Integer
  Dim BackgroundColour As Long
  
  If (LCase$(Left$(Cmnd, 3)) = "rgb") Then
    CmndCpy = Trim$(Mid$(Cmnd, 4))
    Posn = InStr(1, CmndCpy, ",")
    Red = Val(Left$(CmndCpy, (Posn - 1)))
    CmndCpy = Trim$(Mid$(Cmnd, (Posn + 1)))
    Posn = InStr(1, CmndCpy, ",")
    Green = Val(Left$(CmndCpy, (Posn - 1)))
    CmndCpy = Trim$(Mid$(Cmnd, (Posn + 1)))
    Blue = Val(CmndCpy)
    
    ' the order below is Blue, Green, Red as Windows stores
    ' these in the form 0xBBGGRR...
    BackgroundColour = RGB(Blue, Green, Red)
    
    ' If colour is black, make it '1' so we can pick up on
    ' zero being "value was never assigned"; which should not
    ' happen, but...
    If (BackgroundColour = 0&) Then BackgroundColour = 1&
    
    DocVars.Background = BackgroundColour
  
  ElseIf (LCase$(Left$(Cmnd, 4)) = "wimp") Then
    Red = Val(Mid$(Cmnd, 5))
    
    If (Red = 1) Then
      ' WIMP colour 1, enable tiling
      Red = -123
    Else
      ' convert colours 0...15 to range -1...-16
      Red = -1 - Red
    End If
    
    MyStr = MyStr & CStr(Red)
    DocVars.Background = Red
  End If
End Sub


Private Sub Help_ParseFontDef(ByVal Cmnd As String, _
                              ByVal FP As Integer)

  Dim MyStr As String
'  MyStr = "<!-- Font/Style def: " & Chr$(34) & Cmnd & Chr$(34) & _
          " -->" & vbCrLf
'  Put #FP, , MyStr

End Sub


Private Function Help_ReturnBackground() As String
  ' This returns a string defining the "<body>" tag, which has
  ' the current background options set up.
  '
  ' The possible background colours are:
  '   +ve : A 'long' defining an RGB value for the background.
  '     1 : This means "black", because zero has a meaning.
  '     0 : This means "undefined", the value for the background
  '         was NEVER defined, hence we will revert to &EEEEEE
  '         (very light grey) and add tiling (i.e. like Wimp 1).
  '   -ve : Negative -1 to -16 is WIMP colour 0 to 15.
  '  -123 : Special value means "set Wimp colour 1 and use tile".
  '
  
  Dim MyCol As Long, MyStr As String
  
  If (DocVars.Background > 1) Then
    ' +ve, it's an RGB colour
    MyCol = DocVars.Background
    
  ElseIf (DocVars.Background = 1) Then
    ' It is black
    MyCol = 0&
  
  ElseIf (DocVars.Background = 0) Then
    ' It is unset, force to light grey (and add tile)
    MyCol = &HEEEEEE
  
  Else
    ' It is negative, so...
    
    If (DocVars.Background = -123) Then
      ' Want tiled, set background to Wimp colour 1.
      MyCol = WimpColours(1)
    
    Else
      ' Just a regular WIMP colour
      MyCol = WimpColours(Abs(DocVars.Background) - 1)
    End If
  End If
  
  ' Start defining the body
  MyStr = "<body lang=" & Chr$(34) & "en-gb" & Chr$(34)
  
  ' Add in the background colour
  MyStr = MyStr & " bgcolor=" & Chr$(34) & "#" & _
          Right$("000000" & Hex$(MyCol), 6) & Chr$(34)
  
  ' Is there to be a background tile?
  If ((DocVars.Background = -123) Or _
      (DocVars.Background = 0)) Then
    MyStr = MyStr & " background=" & Chr$(34) & "tile.jpeg" & _
            Chr$(34)
  End If
  
  ' Now set the other colours; we set link and vlink to be the
  ' same so that it doesn't show up different for pages that have
  ' been 'visited'.
  MyStr = MyStr & " text=black link=blue vlink=blue " & _
          "alink=navy> " & vbCrLf & vbCrLf

  Help_ReturnBackground = MyStr
End Function


Private Function Help_SortOutFontMess(ByVal FMess As String, _
                                      ByVal LBrk As Boolean) _
                                      As String
  ' Oh, God Almighty, where do we start? Okay...
  '
  '   f
  '     Restore the previous font options.
  '
  '   f<style>
  '     Set the nominated style as current
  '
  '   f<number>
  '     Set the nominated font as current
  '
  '   f<num> = <face> <size> [Bold <n>] [Italic <n>] [Both <n>]
  '     Define the numbered font to be "<face>" at <size> points.
  '     Actually I think there can be a second size for different
  '     vertical and horizontal sizes. OBVIOUSLY that is not going
  '     to be supported - I'm not sure HTML can even do that...?
  '     The <n>umber for Bold/Italic/Both are to set which font
  '     definition is to be used when setting bold/italic/both.
  '     [currently we just specify <b> or <i> as applicable]
  '
  '   f<style> = f[<num>][/][_][*] [rgb <r>,<g>,<b>] [align <align>]
  '     This defines the named style. The parser must be able to
  '     support stuff like "= f_*" which means "this means set
  '     bold and italics".
  '
  '     The handling is fairly rudimentary right now, we have our
  '     own custom style for links... it may work out a lot better
  '     to parse this into a CSS block to whack out at the start,
  '     and then wrap all this crap in "<span class=...>" bits.
  '
  
  Dim FStr As String, OStr As String, Posn As Integer
  Dim MyLine As String
  
  ' Just a turn-off?
  If (FMess = "") Then
    ' Just a plain "f", so unset...
    MyLine = "</font>"
    Help_SortOutFontMess = MyLine
    Exit Function
  End If
  
  ' A definition?
  Posn = InStr(1, FMess, "=")
  If (Posn <> 0) Then Exit Function ' just IGNORE definitions
  
  ' Linked text?
  Posn = InStr(1, FMess, ":")
  If (Posn <> 0) Then
    FStr = Left$(FMess, (Posn - 1))
    OStr = Mid$(FMess, (Posn + 1))
  Else
    FStr = FMess
    OStr = ""
  End If
    
  ' Can we evaluate it?
  If (Val(FStr) <> 0) Then Exit Function ' if so, IGNORE
  
  ' So it is probably a style. Pick up on the default ones
  ' [note to self: shouldn't this be configurable?]
  If (StrComp(FStr, "Std", vbTextCompare) = 0) Then
    MyLine = "<font style=" & Chr$(34) & "font-family: " & _
             "'Times New Roman'; font-size: 12pt;" & Chr$(34) & _
             ">"
  
  ElseIf (StrComp(FStr, "Code", vbTextCompare) = 0) Then
    MyLine = "<font style=" & Chr$(34) & "font-family: " & _
             "'Courier New'; font-size: 12pt;" & Chr$(34) & _
             ">"

  ElseIf (StrComp(FStr, "H1", vbTextCompare) = 0) Then
    MyLine = "<font style=" & Chr$(34) & "font-family: " & _
             "'Arial'; font-weight: bold; font-size: 26pt; " & _
             "text-align: center;" & Chr$(34) & ">"

  ElseIf (StrComp(FStr, "H2", vbTextCompare) = 0) Then
    MyLine = "<font style=" & Chr$(34) & "font-family: " & _
             "'Arial'; font-weight: bold; font-size: 20pt; " & _
             "font-style: oblique; text-align: center;" & _
             Chr$(34) & ">"

  ElseIf (StrComp(FStr, "H3", vbTextCompare) = 0) Then
    MyLine = "<font style=" & Chr$(34) & "font-family: " & _
             "'Times New Roman'; font-weight: bold; " & _
             "font-size: 18pt; text-align: center;" & _
             Chr$(34) & ">"

  ElseIf (StrComp(FStr, "H4", vbTextCompare) = 0) Then
    MyLine = "<font style=" & Chr$(34) & "font-family: " & _
             "'Times New Roman'; font-weight: bold; " & _
             "font-size: 16pt; font-style: oblique; " & _
             "text-align: center;" & Chr$(34) & ">"

  ElseIf (StrComp(FStr, "H5", vbTextCompare) = 0) Then
    MyLine = "<font style=" & Chr$(34) & "font-family: " & _
             "'Times New Roman'; font-weight: bold; " & _
             "font-size: 14pt; text-align: center;" & _
             Chr$(34) & ">"

  ElseIf (StrComp(FStr, "H6", vbTextCompare) = 0) Then
    MyLine = "<font style=" & Chr$(34) & "font-family: " & _
             "'Times New Roman'; font-weight: bold; " & _
             "font-size: 14pt; font-style: oblique; " & _
             "text-align: center;" & Chr$(34) & ">"

  Else
    ' no others supported...
    Debug.Print "Missed on font def " & FStr
  End If
  
  If ((OStr <> "") Or (Left$(FStr, 1) = ";")) Then
    MyLine = MyLine & OStr & "</font>"
    If (LBrk = True) Then
      '     for: #fBlah:Some text
      ' but not: {fBlah:Some text}
      MyLine = MyLine & "<br>" & vbCrLf
    End If
  End If

  Help_SortOutFontMess = MyLine
End Function


Private Function Help_WonkAnEffect(ByVal ThisByte As Byte, _
                                   ByVal PrevByte As Byte, _
                                   ByVal FP As Integer)
  ' Support *wonky* /effects/.
  '
  ' Here's how it works: We have the previous byte, the current
  ' byte, and we'll read in the next byte.
  '
  ' If *one* (and *only* *ONE*) of the prev/next is an alphanumeric
  ' then we will permit the effect. This means that *bold* will
  ' become bold, but diddly/squat will not become part-italicised.
  ' It will also leave alone stand-alone things like 44 / 10 = 4.4
  '
  
  Dim NextByte As Byte, MyPtr As Long, MyStr As String
  Dim PrevIsAlphaNum As Boolean, NextIsAlphaNum As Boolean
  Dim ScanByte As Byte
  
  ' transparently get the next byte (a BytePeek)
  MyPtr = Seek(FP)
  Get #FP, , NextByte
  Seek #FP, MyPtr
  
  ScanByte = PrevByte ' saves messing with the original
  If (ScanByte >= 97) Then ScanByte = ScanByte And -33
  If (NextByte >= 97) Then NextByte = NextByte And -33
  
  PrevIsAlphaNum = IsAlphaNum(ScanByte)
  NextIsAlphaNum = IsAlphaNum(NextByte)
  
  If (PrevIsAlphaNum <> NextIsAlphaNum) Then
    MyStr = Help_ParseCommand(Chr$(ThisByte), FP, False)
  Else
    MyStr = MyStr & Chr$(ThisByte)
  End If
  
  Help_WonkAnEffect = MyStr
End Function


Private Sub Help_CreateTable(ByVal FP As Integer)
  ' Try to make a table...
  '
  ' IMPORTANT:
  '
  '      In StongHelp, tables read DOWNWARDS, not across.
  '      ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
  '         (down, not across...  down, not across...)
  
  Dim TCols As Integer, TRows As Integer, MyLine As String
  Dim CLoop As Integer, RLoop As Integer, TOffs As Integer
  Dim CellWidth As Integer
  
  ' trap failed entry
  If ((TableCols = -1) And (TableRows = -1)) Then Exit Sub
  
  If (TableRows = -1) Then
    ' We're using columns, so...
    TRows = TableCount / TableCols
    TCols = TableCols
    ' To make 'x' columns, we'll need a table with <y> rows...
  Else
    ' We're using rows so...
    TCols = TableCount / TableRows
    TRows = TableRows
    ' To make 'x' rows, we'll need a table with <y> columns...
  End If
  
  CellWidth = 100 \ TCols
  
  MyLine = vbCrLf & vbCrLf & "<table border=0 width=" & _
           Chr$(34) & "100%" & Chr$(34) & ">" & vbCrLf
  Put #FP, , MyLine
  
  For RLoop = 1 To TRows
    MyLine = "  <tr>" & vbCrLf
    Put #FP, , MyLine
    
    For CLoop = 1 To TCols
      If (RLoop = 1) Then
        MyLine = "    <td width=" & Chr$(34) & CStr(CellWidth) & _
                 "%" & Chr$(34) & ">"
      Else
        MyLine = "    <td>"
      End If
      Put #FP, , MyLine
    
      TOffs = RLoop + ((CLoop - 1) * TRows)
      If (TOffs > TableCount) Then
        MyLine = "&nbsp; <!-- no data! --> </td>" & vbCrLf
        Put #FP, , MyLine
      Else
        MyLine = TableData(TOffs) & "</td>" & vbCrLf
        Put #FP, , MyLine
      End If
    Next
    
    MyLine = "  </tr>" & vbCrLf
    Put #FP, , MyLine
  Next
  
  MyLine = "</table>" & vbCrLf & vbCrLf
  Put #FP, , MyLine
  
  TableCount = 0
  ReDim TableData(TableCount)
End Sub


Public Sub Help_PushToStack(ByVal Ref As Integer)
  ' Place a document on the page stack
  
  ' For the MSIE8 double-clicky balls, if this new reference is
  ' the same as the current top reference, just don't...
  If (StackCount > 0) Then
    If (Ref = PageStack(StackCount)) Then Exit Sub
  End If
  ' That is not a problem in use as it is HIGHLY unlikely that
  ' a help page would link to itself, and if it should, we can
  ' be smart and collapse it to a single entry. The benefit,
  ' on the other hand, is the user no longer needs to double
  ' go-back in normal use... which is infinitely worthwhile.
  ' [stupid MSIE8 altering past behaviour like that!]
  
  StackCount = StackCount + 1
  ReDim Preserve PageStack(StackCount)
  PageStack(StackCount) = Ref
  
  If (Ref = ManualRootEntry) Then
    frmMain.btnMain.Enabled = False
  Else
    frmMain.btnMain.Enabled = True
  End If
End Sub


Public Sub Help_PullFromStack()
  ' Remove a document from the page stack, forcing it to reload.
  If (StackCount = 1) Then
    MsgBox "You cannot go back.", vbExclamation + vbOKOnly
    Exit Sub
  End If
  
  StackCount = StackCount - 1
  ReDim Preserve PageStack(StackCount)
  If (StackCount > 0) Then
    Help_ConvertToHTML PageStack(StackCount)
  Else
    MsgBox "You cannot go back any further as there is no " & _
           "more history.", vbExclamation + vbOKOnly
  End If

  If (PageStack(StackCount) = ManualRootEntry) Then
    frmMain.btnMain.Enabled = False
  Else
    frmMain.btnMain.Enabled = True
  End If
End Sub


Public Sub Help_WonkStackTo(ByVal Where As String)
  ' Screw around with the stack
  
  Dim NewLoc As Integer
  
  Unload frmHistory
  
  NewLoc = Val(Where) + 1
  StackCount = NewLoc
  ReDim Preserve PageStack(StackCount)
  Help_PullFromStack
End Sub


Public Sub Help_ShowStack(ByVal xpos As Single, _
                          ByVal ypos As Single)
  ' Display the current history stack...
  
  Dim Which As Integer, TxtHei As Single, Loca As String
  
  Load frmHistory
  
  frmHistory.liHistory.Clear
  For Which = 1 To StackCount
    If (Which = 1) Then
      Loca = "<initial page>"
    Else
      Loca = Files(PageStack(Which)).BRName
      If (StrComp(Right$(Loca, 5), "!root", _
                  vbTextCompare) = 0) Then
        Loca = Left$(Loca, Len(Loca) - 5) & " <base page>"
      End If
    End If
    frmHistory.liHistory.AddItem CStr(Which) & vbTab & Loca
  Next
  
  If (StackCount <= 24) Then
    TxtHei = frmHistory.TextHeight("X") * CSng(StackCount + 1)
  Else
    TxtHei = frmHistory.TextHeight("X") * 25#
  End If
  
  frmHistory.liHistory.Height = TxtHei
  
  frmHistory.Height = frmHistory.liHistory.Height + 105 + 480
  
  frmHistory.Left = xpos - 200
  frmHistory.Top = ypos
    
  frmHistory.Show vbModal
End Sub


Public Sub Help_Discard()
  ' Discard cache, indent levels, etc...
  
  DispDepth = 0
  ' DispDepth is fixed 0-32

  TableCount = 0
  ReDim TableEntries(TableCount)
  InTable = False

  StackCount = 0
  ReDim PageStack(StackCount)

  IndentCount = 0
  ReDim Indents(IndentCount)
  
  Unload frmFileSys
End Sub


Public Function Help_ReturnTypeOfFile(ByVal Which As Integer) _
                                     As Integer
  ' Saves making the file array global. This is for the FileSys
  ' loader to prevent stupid behaviour if, say, trying to load
  ' and parse an SQSH file!
  
  Help_ReturnTypeOfFile = Files(Which).FileType
End Function



Public Sub Help_DumpTables()
  ' For development, this will permit us to see the directory
  ' and file structures.
  
  Dim FP As Integer, ObjRef As Integer, LineOut As String
  Dim FileType As Integer
  
  If (Dir(App.Path & "\tables.txt") <> "") Then
    Kill App.Path & "\tables.txt"
  End If
  
  FP = FreeFile
  Open App.Path & "\tables.txt" For Binary Access Write As FP
  
  LineOut = "DIRECTORIES (" & CStr(DirCount) & ")" & vbCrLf & _
            "===========" & vbCrLf & vbCrLf
  Put #FP, , LineOut
  
  For ObjRef = 1 To DirCount
    LineOut = Right$("    " & CStr(ObjRef), 4) & _
              ": Name   = " & Dirs(ObjRef).Name & vbCrLf & _
              "      FQName = " & Dirs(ObjRef).FQName & _
              vbCrLf & "      Offset = +" & _
              CStr(Dirs(ObjRef).Offset) & vbCrLf & _
              "      Size   = " & CStr(Dirs(ObjRef).Size) & _
              vbCrLf & "      Ld/Exc = " & Right$("00000000" & _
              Hex(Dirs(ObjRef).LoadAddr), 8) & ", " & _
              Right$("00000000" & Hex$(Dirs(ObjRef).ExecAddr), _
              8) & vbCrLf & "      Flags  = " & _
              CStr(Dirs(ObjRef).Flags) & vbCrLf & _
              "      Parent = " & CStr(Dirs(ObjRef).Parent) & _
              vbCrLf & vbCrLf
    Put #FP, , LineOut
  Next
  
  LineOut = vbCrLf & vbCrLf & "FILES (" & CStr(FileCount) & ")" & _
            vbCrLf & "=====" & vbCrLf & vbCrLf
  Put #FP, , LineOut
  
  For ObjRef = 1 To FileCount
    LineOut = Right$("    " & CStr(ObjRef), 4) & _
              ": Name   = " & Files(ObjRef).Name & vbCrLf & _
              "      FQName = " & Files(ObjRef).FQName & _
              vbCrLf & "      BRName = " & Files(ObjRef).BRName & _
              vbCrLf & "      Offset = +" & _
              CStr(Files(ObjRef).Offset) & vbCrLf & _
              "      Size   = " & CStr(Files(ObjRef).Size) & _
              vbCrLf
              FileType = Files(ObjRef).FileType
              If (FileType <> 0) Then
                ' Typed file
                LineOut = LineOut & "      Type   = &" & _
                          Right$("000" & Hex$(FileType), 3)
                Select Case FileType
                  Case &HAFF: LineOut = LineOut & _
                                " (DrawFile)"
                  Case &HFCA: LineOut = LineOut & _
                                " (SQSH compressed data)"
                  Case &HFF9: LineOut = LineOut & _
                                " (sprite image)"
                  Case &HFFC: LineOut = LineOut & _
                                " ('utility' code (for '!pre'))"
                  Case &HFFD: LineOut = LineOut & _
                                " (non-searchable help text)"
                  Case &HFFF: LineOut = LineOut & _
                                " (searchable help text)"
                End Select
                LineOut = LineOut & vbCrLf
                
                ' calculating the date stamp in VB would be an exercise
                ' in pain, and actually it is NOT relevant to us!
              
              Else
                ' Load/Exec addresses
                LineOut = LineOut & "      Ld/Exc = " & _
                Right$("00000000" & Hex(Files(ObjRef).LoadAddr), _
                8) & ", " & Right$("00000000" & _
                Hex$(Files(ObjRef).ExecAddr), 8) & vbCrLf
              End If
              LineOut = LineOut & "      Flags  = " & _
              CStr(Files(ObjRef).Flags) & vbCrLf & _
              "      Parent = " & CStr(Files(ObjRef).Parent) & _
              vbCrLf & vbCrLf
    Put #FP, , LineOut
  Next
  
  LineOut = vbCrLf & vbCrLf & "HISTORY (" & CStr(StackCount) & ")" & _
            vbCrLf & "=======" & vbCrLf & vbCrLf
  Put #FP, , LineOut

  If (StackCount > 0) Then
    For ObjRef = 1 To StackCount
      LineOut = Right$("   " & CStr(ObjRef), 3) & ". " & _
                Files(PageStack(ObjRef)).BRName & vbCrLf
      Put #FP, , LineOut
    Next
  Else
    LineOut = "There are no page history entries." & vbCrLf
    Put #FP, , LineOut
  End If
  
  LineOut = vbCrLf & vbCrLf & vbCrLf & "[end of report]" & vbCrLf
  Put #FP, , LineOut
  
  Close #FP
  
  Shell "Notepad " & Chr$(34) & App.Path & "\tables.txt" & _
        Chr$(34), vbNormalFocus
End Sub
